/*

*/

import renderer.scene.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;
import javax.swing.JFrame;
import java.awt.BorderLayout;
import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;
import java.awt.event.ComponentListener;
import java.awt.event.ComponentEvent;

/**
   Draw an interactive robot arm with shoulder, elbow, wrist, and finger joints.
   Make each part of the robot arm extendable.
<p>
   The tree for this scene is shown below.
<p>
   Remember that every position node in the tree contains a matrix,
   a model and a list of nested positions. The model may be empty,
   and the list of nested positions may also be empty, but the matrix
   cannot be "empty" (if you don't give it a value, then it is the
   identity matrix, I).
<p>
<pre>{@code
          Scene
         /     \
        /       \
  Camera   List<Position>
               |
               |
            Position
            /  |    \
           /   |     \
     Matrix    |     List<Position>
       RS      |           |
               |           |
              /        Position
             /         /   |   \
            /         /    |    \
           /     Matrix   /      List<Position>
          |       TRS    /             |
          |             /              |
          |            /            Position
          |           /             /  |   \
          |          /             /   |    \
          |         /        Matrix    |     List<Position>
          |        /          TRS      |           |
          |       /                   /            |
           \     /                   /          Position
            Model ------------------/           /  |   \
          armSegment                           /   |    \
                    \                    Matrix    |     List<Position>
                     \                    TRS     /             |
                      \                          /            empty
                       \------------------------/

</pre>
*/
public class RobotArm3 implements KeyListener, ComponentListener
{
   private double shoulderRotation = 0.0;
   private double    elbowRotation = 0.0;
   private double    wristRotation = 0.0;
   private double   fingerRotation = 0.0;

   private double shoulderLength = 0.4;
   private double    elbowLength = 0.3;
   private double    wristLength = 0.2;
   private double   fingerLength = 0.1;

   private Scene scene = null;
   private Position shoulder_p = null;
   private Position elbow_p = null;
   private Position wrist_p = null;
   private Position finger_p = null;

   private JFrame jf = null;
   private FrameBufferPanel fbp = null;

   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
      Then this constructor instantiates the GUI.
   */
   public RobotArm3()
   {
      // Create the Scene object that we shall render.
      scene = new Scene();

      // Create a Position object that will hold the robot arm.
      shoulder_p = new Position();

      // Add the Position object to the scene.
      scene.addPosition(shoulder_p);

      // Push the position away from where the camera is.
      shoulder_p.matrix = Matrix.translate(0, 0, -1);

      /*
         Create the scene graph.
      */
      // Create one Model that can be used
      // for each part of the robot arm.
      final Vertex v0 = new Vertex(0, 0, 0);
      final Vertex v1 = new Vertex(1, 0, 0);
      final Model armSegment = new Model("arm segment");
      armSegment.addVertex(v0, v1);
      armSegment.addLineSegment(new LineSegment(0, 1));
      ModelShading.setColor(armSegment, Color.blue);

      // Add the armSegment Model to the Scene's Position.
      shoulder_p.model = armSegment;

      elbow_p = new Position(armSegment);
      shoulder_p.addNestedPosition(elbow_p);

      wrist_p = new Position(armSegment);
      elbow_p.addNestedPosition(wrist_p);

      finger_p = new Position(armSegment);
      wrist_p.addNestedPosition(finger_p);

      // Initialize the nested matrices for the sub models.
      shoulder_p.matrix.mult(Matrix.scale(shoulderLength,
                                          shoulderLength,
                                          1));
      elbow_p.matrix.mult(Matrix.translate(1, 0, 0));
      elbow_p.matrix.mult(Matrix.scale(elbowLength/shoulderLength,
                                       elbowLength/shoulderLength,
                                       1));
      wrist_p.matrix.mult(Matrix.translate(1, 0, 0));
      wrist_p.matrix.mult(Matrix.scale(wristLength/elbowLength,
                                       wristLength/elbowLength,
                                       1));
      finger_p.matrix.mult(Matrix.translate(1, 0, 0));
      finger_p.matrix.mult(Matrix.scale(fingerLength/wristLength,
                                        fingerLength/wristLength,
                                        1));


      // Define initial dimensions for a FrameBuffer.
      final int width  = 1024;
      final int height = 1024;

      // Create a FrameBufferPanel that will hold a FrameBuffer.
      fbp = new FrameBufferPanel(width, height);
      fbp.getFrameBuffer().setBackgroundColorFB(Color.black);
      fbp.getFrameBuffer().getViewport().setBackgroundColorVP(Color.black);

      // Create a JFrame that will hold the FrameBufferPanel.
      jf = new JFrame("Renderer 8");
      jf.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    //jf.setLocationByPlatform(true);
      // Place the FrameBufferPanel in the JFrame.
      jf.getContentPane().add(fbp, BorderLayout.CENTER);
      jf.pack();
      jf.setVisible(true);

      // Create event handler objects for events from the JFrame.
      jf.addKeyListener(this);
      jf.addComponentListener(this);
   }


   // Implement the KeyListener interface.
   @Override public void keyPressed(KeyEvent e)
   {
      //System.out.println( e );

      if (e.isControlDown())
      {
         if (e.isShiftDown())
         {
            if (e.getKeyCode() == KeyEvent.VK_S) // ^S (83)
            {
               shoulderLength -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_E) // ^E (69)
            {
               elbowLength -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_W) // ^W (87)
            {
               wristLength -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_F) // ^F (70)
            {
               fingerLength -= 0.02;
            }
         }
         else // no shift key
         {
            if (e.getKeyCode() == KeyEvent.VK_S) // ^s (93)
            {
               shoulderLength += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_E) // ^e (69)
            {
               elbowLength += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_W)  // ^w (87)
            {
               wristLength += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_F) // ^f (70)
            {
               fingerLength += 0.02;
            }
         }
      }
   }

   @Override public void keyReleased(KeyEvent e){}

   @Override public void keyTyped(KeyEvent e)
   {
      //System.out.println( e );

      final char c = e.getKeyChar();
      if ('h' == c)
      {
         print_help_message();
         return;
      }
      else if ('d' == c)
      {
         Pipeline.debug = ! Pipeline.debug;
         Clip.debug = ! Clip.debug;
      }
      else if ('D' == c)
      {
         RasterizeAntialias.debug = ! RasterizeAntialias.debug;
      }
      else if ('a' == c)
      {
         RasterizeAntialias.doAntialiasing = ! RasterizeAntialias.doAntialiasing;
         System.out.print("Anti-aliasing is turned ");
         System.out.println(RasterizeAntialias.doAntialiasing ? "On" : "Off");
      }
      else if ('g' == c)
      {
         RasterizeAntialias.doGamma = ! RasterizeAntialias.doGamma;
         System.out.print("Gamma correction is turned ");
         System.out.println(RasterizeAntialias.doGamma ? "On" : "Off");
      }
      else if ('c' == c)
      {
         // Change the solid random color of the robot arm.
         Color color = ModelShading.randomColor();
         ModelShading.setColor(shoulder_p.model, color);
         ModelShading.setColor(elbow_p.model, color);
         ModelShading.setColor(wrist_p.model, color);
         ModelShading.setColor(finger_p.model, color);
      }
      else if ('C' == c)
      {
         // Change the solid random color of each segment of the robot arm.
         ModelShading.setRandomColor(shoulder_p.model);
         ModelShading.setRandomColor(elbow_p.model);
         ModelShading.setRandomColor(wrist_p.model);
         ModelShading.setRandomColor(finger_p.model);
      }
      else if ('r' == c)
      {
         // Change the random color at each end of each segment of the robot arm.
         ModelShading.setRainbowLineSegmentColors(shoulder_p.model);
         ModelShading.setRainbowLineSegmentColors(elbow_p.model);
         ModelShading.setRainbowLineSegmentColors(wrist_p.model);
         ModelShading.setRainbowLineSegmentColors(finger_p.model);
      }
      else if ('R' == c)
      {
         // Change the random color at each vertex of the robot arm.
         Color c1 = ModelShading.randomColor();
         Color c2 = ModelShading.randomColor();
         Color c3 = ModelShading.randomColor();
         Color c4 = ModelShading.randomColor();
         Color c5 = ModelShading.randomColor();
         shoulder_p.model.colorList = java.util.Arrays.asList(c1, c2);
            elbow_p.model.colorList = java.util.Arrays.asList(c2, c3);
            wrist_p.model.colorList = java.util.Arrays.asList(c3, c4);
           finger_p.model.colorList = java.util.Arrays.asList(c4, c5);
         shoulder_p.model.getLineSegment(0).setColors(0, 1);
            elbow_p.model.getLineSegment(0).setColors(0, 1);
            wrist_p.model.getLineSegment(0).setColors(0, 1);
           finger_p.model.getLineSegment(0).setColors(0, 1);
      }
      else if ('=' == c)
      {
         shoulderRotation = 0.0;
            elbowRotation = 0.0;
            wristRotation = 0.0;
           fingerRotation = 0.0;

           shoulderLength = 0.4;
              elbowLength = 0.3;
              wristLength = 0.2;
             fingerLength = 0.1;
      }
      else if ('s' == c)
      {
         shoulderRotation += 2.0;
      }
      else if ('S' == c)
      {
         shoulderRotation -= 2.0;
      }
      else if ('e' == c)
      {
         elbowRotation += 2.0;
      }
      else if ('E' == c)
      {
         elbowRotation -= 2.0;
      }
      else if ('w' == c)
      {
         wristRotation += 2.0;
      }
      else if ('W' == c)
      {
         wristRotation -= 2.0;
      }
      else if ('f' == c)
      {
         fingerRotation += 2.0;
      }
      else if ('F' == c)
      {
         fingerRotation -= 2.0;
      }

      // Update the nested matrices for the sub models.
      shoulder_p.matrix2Identity();
      shoulder_p.matrix.mult(Matrix.translate(0, 0, -1));
      shoulder_p.matrix.mult(Matrix.rotateZ(shoulderRotation));
      shoulder_p.matrix.mult(Matrix.scale(shoulderLength,
                                          shoulderLength,
                                          1));

      elbow_p.matrix2Identity();
      elbow_p.matrix.mult(Matrix.translate(1, 0, 0));
      elbow_p.matrix.mult(Matrix.rotateZ(elbowRotation));
      elbow_p.matrix.mult(Matrix.scale(elbowLength/shoulderLength,
                                       elbowLength/shoulderLength,
                                       1));

      wrist_p.matrix2Identity();
      wrist_p.matrix.mult(Matrix.translate(1, 0, 0));
      wrist_p.matrix.mult(Matrix.rotateZ(wristRotation));
      wrist_p.matrix.mult(Matrix.scale(wristLength/elbowLength,
                                       wristLength/elbowLength,
                                       1));

      finger_p.matrix2Identity();
      finger_p.matrix.mult(Matrix.translate(1, 0, 0));
      finger_p.matrix.mult(Matrix.rotateZ(fingerRotation));
      finger_p.matrix.mult(Matrix.scale(fingerLength/wristLength,
                                        fingerLength/wristLength,
                                        1));

      // Render again.
      final FrameBuffer fb = this.fbp.getFrameBuffer();
      fb.clearFB();
      Pipeline.render(scene, fb.vp);
      fbp.update();
   }


   // Implement the ComponentListener interface.
   @Override public void componentMoved(ComponentEvent e){}
   @Override public void componentHidden(ComponentEvent e){}
   @Override public void componentShown(ComponentEvent e){}
   @Override public void componentResized(ComponentEvent e)
   {
      //System.out.println( e );
      /*
      System.out.printf("JFrame [w = %d, h = %d]: " +
                        "FrameBufferPanel [w = %d, h = %d].\n",
                        jf.getWidth(), jf.getHeight(),
                        fbp.getWidth(), fbp.getHeight());
      */
      // Get the new size of the FrameBufferPanel.
      final int w = fbp.getWidth();
      final int h = fbp.getHeight();

      // Create a new FrameBuffer that fits the FrameBufferPanel.
      final Color bg1 = fbp.getFrameBuffer().getBackgroundColorFB();
      final Color bg2 = fbp.getFrameBuffer().getViewport().getBackgroundColorVP();
      final FrameBuffer fb = new FrameBuffer(w, h, bg1);
      fb.vp.setBackgroundColorVP(bg2);
      fbp.setFrameBuffer(fb);
      fb.clearFB();
      Pipeline.render(scene, fb.vp);
      fbp.update();
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      print_help_message();

      // We need to call the program's constructor in the
      // Java GUI Event Dispatch Thread, otherwise we get a
      // race condition between the constructor (running in
      // the main() thread) and the very first ComponentEvent
      // (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         () -> new RobotArm3()
      );
   }


   private static void print_help_message()
   {
      System.out.println("Use the 'd' key to toggle debugging information on and off.");
      System.out.println("Use the 'a' key to toggle antialiasing on and off.");
      System.out.println("Use the 'g' key to toggle gamma correction on and off.");
      System.out.println("Use the 'c' key to change the random solid arm color.");
      System.out.println("Use the 'C' key to randomly change arm segment colors.");
      System.out.println("Use the 'r' key to randomly change arm segment end colors.");
      System.out.println("Use the 'R' key to randomly change arm hinge colors.");
      System.out.println("Use the s/S keys to rotate the arm at the shoulder.");
      System.out.println("Use the e/E keys to rotate the arm at the elbow.");
      System.out.println("Use the w/W keys to rotate the arm at the wrist.");
      System.out.println("Use the f/F keys to rotate the arm at the finger.");
      System.out.println("Use the ^s/^S keys to extend the length of the arm at the shoulder.");
      System.out.println("Use the ^e/^E keys to extend the length of the arm at the elbow.");
      System.out.println("Use the ^w/^W keys to extend the length of the arm at the wrist.");
      System.out.println("Use the ^f/^F keys to extend the length of the arm at the finger.");
      System.out.println("Use the '=' key to reset the robot arm.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }
}
