/*

*/

import renderer.scene.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;
import javax.swing.JFrame;
import java.awt.BorderLayout;
import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;
import java.awt.event.ComponentListener;
import java.awt.event.ComponentEvent;

/**
   Draw an interactive robot arm with one shoulder,
   two elbow, two wrist, and two finger joints.
<p>
   Here is a simplified version of this program's scene graph.
<p>
<pre>{@code
              Scene
                |
                |
            Position
            / | \   \
           /  |  \   \--------------------\
     Matrix   |   \                        \
      TRS    /  Position                 Position
            /     / |  \                  /  |  \
           /     /  |   \                /   |   \
          / Matrix  |    \          Matrix   |    \
         /   TRS   /    Position     TRS    /  Position
         \        /      / |  \            /    / |   \
          \      /      /  |   \          /    /  |    \
           \    /  Matrix  |    \        / Matrix |     \
            \   \   TRS   /   Position  |   TRS   /   Position
             \   \       /      /  |    |        /     /  |
              \   \     /      /   |    |       /     /   |
               \   \   /  Matrix   |    |      /  Matrix  |
                \   \  \   TRS     |    |     /    TRS   /
                 \   \  \          |    |    /          /
                  \   \  \-----\   |    /   /          /
                   \   \--------\  |   /---/          /
                    \------------\ |  /--------------/
                                 Model
                               armSegment
</pre>
*/
public class RobotArms1 implements KeyListener, ComponentListener
{
   private double xTranslation = 0.0;
   private double yTranslation = 0.0;

   private double shoulderRotation =  0.0;
   private double   elbowRotation1 =  15.0;
   private double   elbowRotation2 = -15.0;
   private double   wristRotation1 =  0.0;
   private double   wristRotation2 =  0.0;
   private double  fingerRotation1 =  0.0;
   private double  fingerRotation2 =  0.0;

   private double  shoulderLength = 0.4;
   private double    elbowLength1 = 0.3;
   private double    elbowLength2 = 0.3;
   private double    wristLength1 = 0.2;
   private double    wristLength2 = 0.2;
   private double   fingerLength1 = 0.1;
   private double   fingerLength2 = 0.1;

   private Scene scene = null;
   private Position shoulder_p = null;
   private Position elbow1_p = null;
   private Position elbow2_p = null;
   private Position wrist1_p = null;
   private Position wrist2_p = null;
   private Position finger1_p = null;
   private Position finger2_p = null;

   private JFrame jf = null;
   private FrameBufferPanel fbp = null;

   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
      Then this constructor instantiates the GUI.
   */
   public RobotArms1()
   {
      // Create the Scene object that we shall render.
      scene = new Scene();

      // Create a Position object that will hold the robot arm.
      shoulder_p = new Position();

      // Add the Position object to the scene.
      scene.addPosition(shoulder_p);

      /*
         Create the scene graph.
      */
      // Create one Model that can be used
      // for each part of the robot arms.
      final Vertex v0 = new Vertex(0, 0, 0);
      final Vertex v1 = new Vertex(1, 0, 0);
      final Model armSegment = new Model("arm segment");
      armSegment.addVertex(v0, v1);
      armSegment.addLineSegment(new LineSegment(0, 1));
      ModelShading.setColor(armSegment, Color.blue);

      // Add the armSegment Model to the Scene's Position.
      shoulder_p.model = armSegment;

      // two elbows
      elbow1_p = new Position(armSegment);
      elbow2_p = new Position(armSegment);
      shoulder_p.addNestedPosition(elbow1_p);
      shoulder_p.addNestedPosition(elbow2_p);

      // two wrists
      wrist1_p = new Position(armSegment);
      wrist2_p = new Position(armSegment);
      elbow1_p.addNestedPosition(wrist1_p);
      elbow2_p.addNestedPosition(wrist2_p);

      // two fingers
      finger1_p = new Position(armSegment);
      finger2_p = new Position(armSegment);
      wrist1_p.addNestedPosition(finger1_p);
      wrist2_p.addNestedPosition(finger2_p);

      // Initialize the nested matrices for the sub models.
      shoulder_p.matrix2Identity();
      shoulder_p.matrix.mult(Matrix.translate(xTranslation,
                                              yTranslation,
                                              -1));
      shoulder_p.matrix.mult(Matrix.rotateZ(shoulderRotation));
      shoulder_p.matrix.mult(Matrix.scale(shoulderLength,
                                          shoulderLength,
                                          1));
      elbow1_p.matrix2Identity();
      elbow1_p.matrix.mult(Matrix.translate(1, 0, 0));
      elbow1_p.matrix.mult(Matrix.rotateZ(elbowRotation1));
      elbow1_p.matrix.mult(Matrix.scale(elbowLength1/shoulderLength,
                                        elbowLength1/shoulderLength,
                                        1));
      elbow2_p.matrix2Identity();
      elbow2_p.matrix.mult(Matrix.translate(1, 0, 0));
      elbow2_p.matrix.mult(Matrix.rotateZ(elbowRotation2));
      elbow2_p.matrix.mult(Matrix.scale(elbowLength2/shoulderLength,
                                        elbowLength2/shoulderLength,
                                        1));
      wrist1_p.matrix2Identity();
      wrist1_p.matrix.mult(Matrix.translate(1, 0, 0));
      wrist1_p.matrix.mult(Matrix.rotateZ(wristRotation1));
      wrist1_p.matrix.mult(Matrix.scale(wristLength1/elbowLength1,
                                        wristLength1/elbowLength1,
                                        1));
      wrist2_p.matrix2Identity();
      wrist2_p.matrix.mult(Matrix.translate(1, 0, 0));
      wrist2_p.matrix.mult(Matrix.rotateZ(wristRotation2));
      wrist2_p.matrix.mult(Matrix.scale(wristLength2/elbowLength2,
                                        wristLength2/elbowLength2,
                                        1));
      finger1_p.matrix2Identity();
      finger1_p.matrix.mult(Matrix.translate(1, 0, 0));
      finger1_p.matrix.mult(Matrix.rotateZ(fingerRotation1));
      finger1_p.matrix.mult(Matrix.scale(fingerLength1/wristLength1,
                                         fingerLength1/wristLength1,
                                         1));
      finger2_p.matrix2Identity();
      finger2_p.matrix.mult(Matrix.translate(1, 0, 0));
      finger2_p.matrix.mult(Matrix.rotateZ(fingerRotation2));
      finger2_p.matrix.mult(Matrix.scale(fingerLength2/wristLength2,
                                         fingerLength2/wristLength2,
                                         1));


      // Define initial dimensions for a FrameBuffer.
      final int width  = 1024;
      final int height = 1024;

      // Create a FrameBufferPanel that will hold a FrameBuffer.
      fbp = new FrameBufferPanel(width, height);
      fbp.getFrameBuffer().setBackgroundColorFB(Color.black);
      fbp.getFrameBuffer().getViewport().setBackgroundColorVP(Color.black);

      // Create a JFrame that will hold the FrameBufferPanel.
      jf = new JFrame("Renderer 8");
      jf.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    //jf.setLocationByPlatform(true);
      // Place the FrameBufferPanel in the JFrame.
      jf.getContentPane().add(fbp, BorderLayout.CENTER);
      jf.pack();
      jf.setVisible(true);

      // Create event handler objects for events from the JFrame.
      jf.addKeyListener(this);
      jf.addComponentListener(this);
   }


   // Implement the KeyListener interface.
   @Override public void keyPressed(KeyEvent e)
   {
      //System.out.println( e );

      if (e.isControlDown())
      {
         if (e.isShiftDown())
         {
            if (e.getKeyCode() == KeyEvent.VK_S) // ^S (83)
            {
               shoulderLength -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_E) // ^E (69)
            {
               elbowLength1 -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_W) // ^W (87)
            {
               wristLength1 -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_F) // ^F (70)
            {
               fingerLength1 -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_Q) // ^Q
            {
               elbowLength2 -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_Z) // ^Z
            {
               wristLength2 -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_D) // ^D
            {
               fingerLength2 -= 0.02;
            }
         }
         else // no shift key
         {
            if (e.getKeyCode() == KeyEvent.VK_S) // ^s (93)
            {
               shoulderLength += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_E) // ^e (69)
            {
               elbowLength1 += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_W)  // ^w (87)
            {
               wristLength1 += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_F) // ^f (70)
            {
               fingerLength1 += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_Q) // ^q
            {
               elbowLength2 += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_Z)  // ^z
            {
               wristLength2 += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_D) // ^d
            {
               fingerLength2 += 0.02;
            }
         }
      }
   }

   @Override public void keyReleased(KeyEvent e){}

   @Override public void keyTyped(KeyEvent e)
   {
      //System.out.println( e );

      final char c = e.getKeyChar();
      if ('h' == c)
      {
         print_help_message();
         return;
      }
      //else if ('d' == c)
      //{
      //   Pipeline.debug = ! Pipeline.debug;
      //   Clip.debug = ! Clip.debug;
      //}
      //else if ('D' == c)
      //{
      //   RasterizeAntialias.debug = ! RasterizeAntialias.debug;
      //}
      //else if ('a' == c)
      //{
      //   RasterizeAntialias.doAntialiasing = ! RasterizeAntialias.doAntialiasing;
      //   System.out.print("Anti-aliasing is turned ");
      //   System.out.println(RasterizeAntialias.doAntialiasing ? "On" : "Off");
      //}
      //else if ('g' == c)
      //{
      //   RasterizeAntialias.doGamma = ! RasterizeAntialias.doGamma;
      //   System.out.print("Gamma correction is turned ");
      //   System.out.println(RasterizeAntialias.doGamma ? "On" : "Off");
      //}
      else if ('c' == c)
      {
         // Change the solid random color of the robot arm.
         Color color = ModelShading.randomColor();
         ModelShading.setColor(shoulder_p.model, color);
         ModelShading.setColor(elbow1_p.model, color);
         ModelShading.setColor(elbow2_p.model, color);
         ModelShading.setColor(wrist1_p.model, color);
         ModelShading.setColor(wrist2_p.model, color);
         ModelShading.setColor(finger1_p.model, color);
         ModelShading.setColor(finger2_p.model, color);
      }
      else if ('C' == c)
      {
         // Change the solid random color of each segment of the robot arm.
         ModelShading.setRandomColor(shoulder_p.model);
         ModelShading.setRandomColor(elbow1_p.model);
         ModelShading.setRandomColor(elbow2_p.model);
         ModelShading.setRandomColor(wrist1_p.model);
         ModelShading.setRandomColor(wrist2_p.model);
         ModelShading.setRandomColor(finger1_p.model);
         ModelShading.setRandomColor(finger2_p.model);
      }
      else if ('r' == c)
      {
         // Change the random color at each end of each segment of the robot arm.
         ModelShading.setRainbowLineSegmentColors(shoulder_p.model);
         ModelShading.setRainbowLineSegmentColors(elbow1_p.model);
         ModelShading.setRainbowLineSegmentColors(elbow2_p.model);
         ModelShading.setRainbowLineSegmentColors(wrist1_p.model);
         ModelShading.setRainbowLineSegmentColors(wrist2_p.model);
         ModelShading.setRainbowLineSegmentColors(finger1_p.model);
         ModelShading.setRainbowLineSegmentColors(finger2_p.model);
      }
      else if ('R' == c)
      {
         // Change the random color at each vertex of the robot arm.
         Color c1 = ModelShading.randomColor();
         Color c2 = ModelShading.randomColor();
         Color c3 = ModelShading.randomColor();
         Color c4 = ModelShading.randomColor();
         Color c5 = ModelShading.randomColor();
         Color c6 = ModelShading.randomColor();
         Color c7 = ModelShading.randomColor();
         Color c8 = ModelShading.randomColor();
         shoulder_p.model.colorList = java.util.Arrays.asList(c1, c2);
           elbow1_p.model.colorList = java.util.Arrays.asList(c2, c3);
           elbow2_p.model.colorList = java.util.Arrays.asList(c2, c4);
           wrist1_p.model.colorList = java.util.Arrays.asList(c3, c5);
           wrist2_p.model.colorList = java.util.Arrays.asList(c4, c6);
          finger1_p.model.colorList = java.util.Arrays.asList(c5, c7);
          finger2_p.model.colorList = java.util.Arrays.asList(c6, c8);
         shoulder_p.model.getLineSegment(0).setColors(0, 1);
           elbow1_p.model.getLineSegment(0).setColors(0, 1);
           elbow2_p.model.getLineSegment(0).setColors(0, 1);
           wrist1_p.model.getLineSegment(0).setColors(0, 1);
           wrist2_p.model.getLineSegment(0).setColors(0, 1);
          finger1_p.model.getLineSegment(0).setColors(0, 1);
          finger2_p.model.getLineSegment(0).setColors(0, 1);
      }
      else if ('=' == c)
      {
         xTranslation = 0.0;
         yTranslation = 0.0;

         shoulderRotation = 0.0;
           elbowRotation1 =  15.0;
           elbowRotation2 = -15.0;
           wristRotation1 = 0.0;
           wristRotation2 = 0.0;
          fingerRotation1 = 0.0;
          fingerRotation2 = 0.0;

           shoulderLength = 0.4;
             elbowLength1 = 0.3;
             elbowLength2 = 0.3;
             wristLength1 = 0.2;
             wristLength2 = 0.2;
            fingerLength1 = 0.1;
            fingerLength2 = 0.1;
      }
      else if ('x' == c)
      {
         xTranslation += 0.02;
      }
      else if ('X' == c)
      {
         xTranslation -= 0.02;
      }
      else if ('y' == c)
      {
         yTranslation += 0.02;
      }
      else if ('Y' == c)
      {
         yTranslation -= 0.02;
      }
      else if ('s' == c)
      {
         shoulderRotation += 2.0;
      }
      else if ('S' == c)
      {
         shoulderRotation -= 2.0;
      }
      else if ('e' == c)
      {
         elbowRotation1 += 2.0;
      }
      else if ('E' == c)
      {
         elbowRotation1 -= 2.0;
      }
      else if ('w' == c)
      {
         wristRotation1 += 2.0;
      }
      else if ('W' == c)
      {
         wristRotation1 -= 2.0;
      }
      else if ('f' == c)
      {
         fingerRotation1 += 2.0;
      }
      else if ('F' == c)
      {
         fingerRotation1 -= 2.0;
      }
      else if ('q' == c)
      {
         elbowRotation2 += 2.0;
      }
      else if ('Q' == c)
      {
         elbowRotation2 -= 2.0;
      }
      else if ('z' == c)
      {
         wristRotation2 += 2.0;
      }
      else if ('Z' == c)
      {
         wristRotation2 -= 2.0;
      }
      else if ('d' == c)
      {
         fingerRotation2 += 2.0;
      }
      else if ('D' == c)
      {
         fingerRotation2 -= 2.0;
      }

      // Set the model matrices for the sub models.
      shoulder_p.matrix2Identity();
      shoulder_p.matrix.mult(Matrix.translate(xTranslation,
                                              yTranslation,
                                              -1));
      shoulder_p.matrix.mult(Matrix.rotateZ(shoulderRotation));
      shoulder_p.matrix.mult(Matrix.scale(shoulderLength,
                                          shoulderLength,
                                          1));

      elbow1_p.matrix2Identity();
      elbow1_p.matrix.mult(Matrix.translate(1, 0, 0));
      elbow1_p.matrix.mult(Matrix.rotateZ(elbowRotation1));
      elbow1_p.matrix.mult(Matrix.scale(elbowLength1/shoulderLength,
                                        elbowLength1/shoulderLength,
                                        1));

      elbow2_p.matrix2Identity();
      elbow2_p.matrix.mult(Matrix.translate(1, 0, 0));
      elbow2_p.matrix.mult(Matrix.rotateZ(elbowRotation2));
      elbow2_p.matrix.mult(Matrix.scale(elbowLength2/shoulderLength,
                                        elbowLength2/shoulderLength,
                                        1));

      wrist1_p.matrix2Identity();
      wrist1_p.matrix.mult(Matrix.translate(1, 0, 0));
      wrist1_p.matrix.mult(Matrix.rotateZ(wristRotation1));
      wrist1_p.matrix.mult(Matrix.scale(wristLength1/elbowLength1,
                                        wristLength1/elbowLength1,
                                        1));

      wrist2_p.matrix2Identity();
      wrist2_p.matrix.mult(Matrix.translate(1, 0, 0));
      wrist2_p.matrix.mult(Matrix.rotateZ(wristRotation2));
      wrist2_p.matrix.mult(Matrix.scale(wristLength2/elbowLength2,
                                        wristLength2/elbowLength2,
                                        1));

      finger1_p.matrix2Identity();
      finger1_p.matrix.mult(Matrix.translate(1, 0, 0));
      finger1_p.matrix.mult(Matrix.rotateZ(fingerRotation1));
      finger1_p.matrix.mult(Matrix.scale(fingerLength1/wristLength1,
                                         fingerLength1/wristLength1,
                                         1));

      finger2_p.matrix2Identity();
      finger2_p.matrix.mult(Matrix.translate(1, 0, 0));
      finger2_p.matrix.mult(Matrix.rotateZ(fingerRotation2));
      finger2_p.matrix.mult(Matrix.scale(fingerLength2/wristLength2,
                                         fingerLength2/wristLength2,
                                         1));


      // Render again.
      final FrameBuffer fb = this.fbp.getFrameBuffer();
      fb.clearFB();
      Pipeline.render(scene, fb.vp);
      fbp.update();
   }


   // Implement the ComponentListener interface.
   @Override public void componentMoved(ComponentEvent e){}
   @Override public void componentHidden(ComponentEvent e){}
   @Override public void componentShown(ComponentEvent e){}
   @Override public void componentResized(ComponentEvent e)
   {
      //System.out.println( e );
      /*
      System.out.printf("JFrame [w = %d, h = %d]: " +
                        "FrameBufferPanel [w = %d, h = %d].\n",
                        jf.getWidth(), jf.getHeight(),
                        fbp.getWidth(), fbp.getHeight());
      */
      // Get the new size of the FrameBufferPanel.
      final int w = fbp.getWidth();
      final int h = fbp.getHeight();

      // Create a new FrameBuffer that fits the FrameBufferPanel.
      final Color bg1 = fbp.getFrameBuffer().getBackgroundColorFB();
      final Color bg2 = fbp.getFrameBuffer().getViewport().getBackgroundColorVP();
      final FrameBuffer fb = new FrameBuffer(w, h, bg1);
      fb.vp.setBackgroundColorVP(bg2);
      fbp.setFrameBuffer(fb);
      fb.clearFB();
      Pipeline.render(scene, fb.vp);
      fbp.update();
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      print_help_message();

      // We need to call the program's constructor in the
      // Java GUI Event Dispatch Thread, otherwise we get a
      // race condition between the constructor (running in
      // the main() thread) and the very first ComponentEvent
      // (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         () -> new RobotArms1()
      );
   }


   private static void print_help_message()
   {
      //System.out.println("Use the 'd' key to toggle debugging information on and off.");
      //System.out.println("Use the 'a' key to toggle antialiasing on and off.");
      //System.out.println("Use the 'g' key to toggle gamma correction on and off.");
      System.out.println("Use the 'c' key to change the random solid arm color.");
      System.out.println("Use the 'C' key to randomly change arm segment colors.");
      System.out.println("Use the 'r' key to randomly change arm segment end colors.");
      System.out.println("Use the 'R' key to randomly change arm hinge colors.");
      System.out.println();
      System.out.println("Use the s/S keys to rotate the arm at the shoulder.");
      System.out.println();
      System.out.println("Use the e/E keys to rotate the arm at elbow 1.");
      System.out.println("Use the w/W keys to rotate the arm at wrist 1.");
      System.out.println("Use the f/F keys to rotate the arm at finger 1.");
      System.out.println();
      System.out.println("Use the q/Q keys to rotate the arm at elbow 2.");
      System.out.println("Use the z/Z keys to rotate the arm at wrist 2.");
      System.out.println("Use the d/D keys to rotate the arm at finger 2.");
      System.out.println();
      System.out.println("Use the ^s/^S keys to extend the length of the arm at the shoulder.");
      System.out.println();
      System.out.println("Use the ^e/^E keys to extend the length of the arm at elbow 1.");
      System.out.println("Use the ^w/^W keys to extend the length of the arm at wrist 1.");
      System.out.println("Use the ^f/^F keys to extend the length of the arm at finger 1.");
      System.out.println();
      System.out.println("Use the ^q/^Q keys to extend the length of the arm at elbow 2.");
      System.out.println("Use the ^z/^Z keys to extend the length of the arm at wrist 2.");
      System.out.println("Use the ^d/^D keys to extend the length of the arm at finger 2.");
      System.out.println();
      System.out.println("Use the x/X keys to translate the arm along the x-axis.");
      System.out.println("Use the y/Y keys to translate the arm along the y-axis.");
      System.out.println();
      System.out.println("Use the '=' key to reset the robot arm.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }
}
