/*

*/

import renderer.scene.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;
import javax.swing.JFrame;
import java.awt.BorderLayout;
import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;
import java.awt.event.ComponentListener;
import java.awt.event.ComponentEvent;

/**
   Draw two interactive robot arms each with one shoulder,
   two elbow, two wrist, and two finger joints.
<p>
   Here is a simplified version of this program's scene graph.
<p>
<pre>{@code
                               Scene
                               /   \
                 /------------/     \-----------------------\
                /                                            \
        Position                                              Position
        / | \   \                                             / | \   \
       /  |  \   \--------------------\                      /  |  \   \--------------------\
 Matrix   /   \                        \               Matrix  /    \                        \
  TRS    /  Position                 Position           TRS   /  Position                 Position
        /     / |  \                  /  |  \                /     / |  \                  /  |  \
       /     /  |   \                /   |   \              /     /  |   \                /   |   \
      / Matrix  |    \          Matrix   |    \            / Matrix  |    \          Matrix   |    \
     |   TRS   /    Position     TRS    /  Position       |   TRS   /    Position     TRS    /  Position
     |        /      / |  \            /    / |   \       |        /      / |  \            /    / |   \
     |       /      /  |   \          /    /  |    \      |       /      /  |   \          /    /  |    \
      \     /  Matrix  |    \        / Matrix |     \     |      |  Matrix  |    \        / Matrix |     \
       \   |    TRS   /   Position  |   TRS   /  Position |      |   TRS   /   Position  |   TRS   /  Position
        \   \        /      /  |    |        /     /    | |      |        /      /  |    |        /     /  |
         \   \      /      /   |    |       |     /     | |      |       /      /   |    |       /     /   |
          \   \    |  Matrix   |    |       | Matrix    | |      |      /  Matrix   |    |      /  Matrix  |
           \   \    \  TRS     |    |       |  TRS      | |      |     /    TRS     |    |     /    TRS   /
            \   \    \         |    |       |           | |      |    /             /    |    /          /
             \   \    \        |    |       \----\      | |      /   /             /     /   /          /
              \   \    \       |    \-------------\     | |     /---/             /     /   /          /
               \   \    \      \-------------------\    | |    /-----------------/     /   /          /
                \   \    \--------------------------\   | |   /-----------------------/   /          /
                 \   \-------------------------------\  | |  /---------------------------/          /
                  \-----------------------------------\ | | /--------------------------------------/
                                                       Model
                                                     armSegment
</pre>
*/
public class RobotArms3 implements KeyListener, ComponentListener
{
   private double[] xTranslation = {0.0,  0.0};
   private double[] yTranslation = {0.5, -0.5};

   private double[] shoulderRotation = {0.0, 0.0};
   private double[]   elbowRotation1 = { 15,  15};
   private double[]   elbowRotation2 = {-15, -15};
   private double[]   wristRotation1 = {0.0, 0.0};
   private double[]   wristRotation2 = {0.0, 0.0};
   private double[]  fingerRotation1 = {0.0, 0.0};
   private double[]  fingerRotation2 = {0.0, 0.0};

   private double[] shoulderLength = {0.4, 0.4};
   private double[]   elbowLength1 = {0.3, 0.3};
   private double[]   elbowLength2 = {0.3, 0.3};
   private double[]   wristLength1 = {0.2, 0.2};
   private double[]   wristLength2 = {0.2, 0.2};
   private double[]  fingerLength1 = {0.1, 0.1};
   private double[]  fingerLength2 = {0.1, 0.1};

   private Scene scene = null;
   private Position[] arm_p = null;
   private Position[] elbow1_p = null;
   private Position[] elbow2_p = null;
   private Position[] wrist1_p = null;
   private Position[] wrist2_p = null;
   private Position[] finger1_p = null;
   private Position[] finger2_p = null;
   private int currentArm = 0;

   private JFrame jf = null;
   private FrameBufferPanel fbp = null;

   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
      Then this constructor instantiates the GUI.
   */
   public RobotArms3()
   {
      // Create the Scene object that we shall render.
      scene = new Scene();

      /*
         Create the scene graph.
      */
      // Create two Position objects that will each hold a robot arm.
      final Position arm1_s = new Position();
      final Position arm2_s = new Position();

      // Add the Position objects to the scene.
      scene.addPosition(arm1_s, arm2_s);

      // Create one Model that can be used
      // for each part of the robot arms.
      final Vertex v0 = new Vertex(0, 0, 0);
      final Vertex v1 = new Vertex(1, 0, 0);
      final Model armSegment = new Model();
      armSegment.addVertex(v0, v1);
      armSegment.addLineSegment(new LineSegment(0, 1));
      ModelShading.setColor(armSegment, Color.blue);

      /*
         Create two robot arms.
      */
      // First arm.
      arm1_s.model = armSegment;

      // two elbows
      final Position arm1_e1 = new Position(armSegment);
      final Position arm1_e2 = new Position(armSegment);
      arm1_s.addNestedPosition(arm1_e1);
      arm1_s.addNestedPosition(arm1_e2);

      // two wrists
      final Position arm1_w1 = new Position(armSegment);
      final Position arm1_w2 = new Position(armSegment);
      arm1_e1.addNestedPosition(arm1_w1);
      arm1_e2.addNestedPosition(arm1_w2);

      // two fingers
      final Position arm1_f1 = new Position(armSegment);
      final Position arm1_f2 = new Position(armSegment);
      arm1_w1.addNestedPosition(arm1_f1);
      arm1_w2.addNestedPosition(arm1_f2);

      // Second arm.
      arm2_s.model = armSegment;

      // two elbows
      final Position arm2_e1 = new Position(armSegment);
      final Position arm2_e2 = new Position(armSegment);
      arm2_s.addNestedPosition(arm2_e1);
      arm2_s.addNestedPosition(arm2_e2);

      // two wrists
      final Position arm2_w1 = new Position(armSegment);
      final Position arm2_w2 = new Position(armSegment);
      arm2_e1.addNestedPosition(arm2_w1);
      arm2_e2.addNestedPosition(arm2_w2);

      // two fingers
      final Position arm2_f1 = new Position(armSegment);
      final Position arm2_f2 = new Position(armSegment);
      arm2_w1.addNestedPosition(arm2_f1);
      arm2_w2.addNestedPosition(arm2_f2);

      arm_p     = new Position[] {arm1_s,  arm2_s};
      elbow1_p  = new Position[] {arm1_e1, arm2_e1};
      elbow2_p  = new Position[] {arm1_e2, arm2_e2};
      wrist1_p  = new Position[] {arm1_w1, arm2_w1};
      wrist2_p  = new Position[] {arm1_w2, arm2_w2};
      finger1_p = new Position[] {arm1_f1, arm2_f1};
      finger2_p = new Position[] {arm1_f2, arm2_f2};

      // Initialize the nested matrices for the sub models.
      // First arm.
      arm_p[0].matrix.mult(Matrix.translate(xTranslation[0],
                                            yTranslation[0],
                                            -1));
      arm_p[0].matrix.mult(Matrix.scale(shoulderLength[0],
                                        shoulderLength[0],
                                        1));
      elbow1_p[0].matrix.mult(Matrix.translate(1, 0, 0));
      elbow1_p[0].matrix.mult(Matrix.rotateZ(elbowRotation1[0]));
      elbow1_p[0].matrix.mult(Matrix.scale(elbowLength1[0]/shoulderLength[0],
                                           elbowLength1[0]/shoulderLength[0],
                                           1));
      elbow2_p[0].matrix.mult(Matrix.translate(1, 0, 0));
      elbow2_p[0].matrix.mult(Matrix.rotateZ(elbowRotation2[0]));
      elbow2_p[0].matrix.mult(Matrix.scale(elbowLength2[0]/shoulderLength[0],
                                           elbowLength2[0]/shoulderLength[0],
                                           1));
      wrist1_p[0].matrix.mult(Matrix.translate(1, 0, 0));
      wrist1_p[0].matrix.mult(Matrix.scale(wristLength1[0]/elbowLength1[0],
                                           wristLength1[0]/elbowLength1[0],
                                           1));
      wrist2_p[0].matrix.mult(Matrix.translate(1, 0, 0));
      wrist2_p[0].matrix.mult(Matrix.scale(wristLength2[0]/elbowLength2[0],
                                           wristLength2[0]/elbowLength2[0],
                                           1));
      finger1_p[0].matrix.mult(Matrix.translate(1, 0, 0));
      finger1_p[0].matrix.mult(Matrix.scale(fingerLength1[0]/wristLength1[0],
                                            fingerLength1[0]/wristLength1[0],
                                            1));
      finger2_p[0].matrix.mult(Matrix.translate(1, 0, 0));
      finger2_p[0].matrix.mult(Matrix.scale(fingerLength2[0]/wristLength2[0],
                                            fingerLength2[0]/wristLength2[0],
                                            1));
      // Second arm.
      arm_p[1].matrix.mult(Matrix.translate(xTranslation[1],
                                            yTranslation[1],
                                            -1));
      arm_p[1].matrix.mult(Matrix.scale(shoulderLength[1],
                                        shoulderLength[1],
                                        1));
      elbow1_p[1].matrix.mult(Matrix.translate(1, 0, 0));
      elbow1_p[1].matrix.mult(Matrix.rotateZ(elbowRotation1[1]));
      elbow1_p[1].matrix.mult(Matrix.scale(elbowLength1[1]/shoulderLength[1],
                                           elbowLength1[1]/shoulderLength[1],
                                           1));
      elbow2_p[1].matrix.mult(Matrix.translate(1, 0, 0));
      elbow2_p[1].matrix.mult(Matrix.rotateZ(elbowRotation2[1]));
      elbow2_p[1].matrix.mult(Matrix.scale(elbowLength2[1]/shoulderLength[1],
                                           elbowLength2[1]/shoulderLength[1],
                                           1));
      wrist1_p[1].matrix.mult(Matrix.translate(1, 0, 0));
      wrist1_p[1].matrix.mult(Matrix.scale(wristLength1[1]/elbowLength1[1],
                                           wristLength1[1]/elbowLength1[1],
                                           1));
      wrist2_p[1].matrix.mult(Matrix.translate(1, 0, 0));
      wrist2_p[1].matrix.mult(Matrix.scale(wristLength2[1]/elbowLength2[1],
                                           wristLength2[1]/elbowLength2[1],
                                           1));
      finger1_p[1].matrix.mult(Matrix.translate(1, 0, 0));
      finger1_p[1].matrix.mult(Matrix.scale(fingerLength1[1]/wristLength1[1],
                                            fingerLength1[1]/wristLength1[1],
                                            1));
      finger2_p[1].matrix.mult(Matrix.translate(1, 0, 0));
      finger2_p[1].matrix.mult(Matrix.scale(fingerLength2[1]/wristLength2[1],
                                            fingerLength2[1]/wristLength2[1],
                                            1));


      // Define initial dimensions for a FrameBuffer.
      final int width  = 1024;
      final int height = 1024;

      // Create a FrameBufferPanel that will hold a FrameBuffer.
      fbp = new FrameBufferPanel(width, height);
      fbp.getFrameBuffer().setBackgroundColorFB(Color.black);
      fbp.getFrameBuffer().getViewport().setBackgroundColorVP(Color.black);

      // Create a JFrame that will hold the FrameBufferPanel.
      jf = new JFrame("Renderer 8");
      jf.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    //jf.setLocationByPlatform(true);
      // Place the FrameBufferPanel in the JFrame.
      jf.getContentPane().add(fbp, BorderLayout.CENTER);
      jf.pack();
      jf.setVisible(true);

      // Create event handler objects for events from the JFrame.
      jf.addKeyListener(this);
      jf.addComponentListener(this);
   }


   // Implement the KeyListener interface.
   @Override public void keyPressed(KeyEvent e)
   {
      //System.out.println( e );

      if (e.isControlDown())
      {
         if (e.isShiftDown())
         {
            if (e.getKeyCode() == KeyEvent.VK_S) // ^S (83)
            {
               shoulderLength[currentArm] -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_E) // ^E (69)
            {
               elbowLength1[currentArm] -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_W) // ^W (87)
            {
               wristLength1[currentArm] -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_F) // ^F (70)
            {
               fingerLength1[currentArm] -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_Q) // ^Q
            {
               elbowLength2[currentArm] -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_Z) // ^Z
            {
               wristLength2[currentArm] -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_D) // ^D
            {
               fingerLength2[currentArm] -= 0.02;
            }
         }
         else // no shift key
         {
            if (e.getKeyCode() == KeyEvent.VK_S) // ^s (93)
            {
               shoulderLength[currentArm] += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_E) // ^e (69)
            {
               elbowLength1[currentArm] += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_W)  // ^w (87)
            {
               wristLength1[currentArm] += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_F) // ^f (70)
            {
               fingerLength1[currentArm] += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_Q) // ^q
            {
               elbowLength2[currentArm] += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_Z)  // ^z
            {
               wristLength2[currentArm] += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_D) // ^d
            {
               fingerLength2[currentArm] += 0.02;
            }
         }
      }
   }

   @Override public void keyReleased(KeyEvent e){}

   @Override public void keyTyped(KeyEvent e)
   {
      //System.out.println( e );

      final char c = e.getKeyChar();
      if ('h' == c)
      {
         print_help_message();
         return;
      }
      //else if ('d' == c)
      //{
      //   Pipeline.debug = ! Pipeline.debug;
      //   Clip.debug = ! Clip.debug;
      //}
      //else if ('D' == c)
      //{
      //   RasterizeAntialias.debug = ! RasterizeAntialias.debug;
      //}
      //else if ('a' == c)
      //{
      //   RasterizeAntialias.doAntialiasing = ! RasterizeAntialias.doAntialiasing;
      //   System.out.print("Anti-aliasing is turned ");
      //   System.out.println(RasterizeAntialias.doAntialiasing ? "On" : "Off");
      //}
      //else if ('g' == c)
      //{
      //   RasterizeAntialias.doGamma = ! RasterizeAntialias.doGamma;
      //   System.out.print("Gamma correction is turned ");
      //   System.out.println(RasterizeAntialias.doGamma ? "On" : "Off");
      //}
      else if ('/' == c)
      {
         currentArm = (currentArm + 1) % 2;
      }
      else if ('c' == c)
      {
         // Change the solid random color of the robot arm.
         Color color = ModelShading.randomColor();
         ModelShading.setColor(    arm_p[currentArm].model, color);
         ModelShading.setColor( elbow1_p[currentArm].model, color);
         ModelShading.setColor( elbow2_p[currentArm].model, color);
         ModelShading.setColor( wrist1_p[currentArm].model, color);
         ModelShading.setColor( wrist2_p[currentArm].model, color);
         ModelShading.setColor(finger1_p[currentArm].model, color);
         ModelShading.setColor(finger2_p[currentArm].model, color);
      }
      else if ('C' == c)
      {
         // Change the solid random color of each segment of the robot arm.
         ModelShading.setRandomColor(    arm_p[currentArm].model);
         ModelShading.setRandomColor( elbow1_p[currentArm].model);
         ModelShading.setRandomColor( elbow2_p[currentArm].model);
         ModelShading.setRandomColor( wrist1_p[currentArm].model);
         ModelShading.setRandomColor( wrist2_p[currentArm].model);
         ModelShading.setRandomColor(finger1_p[currentArm].model);
         ModelShading.setRandomColor(finger2_p[currentArm].model);
      }
      else if ('r' == c)
      {
         // Change the random color at each end of each segment of the robot arm.
         ModelShading.setRainbowLineSegmentColors(arm_p[currentArm].model);
         ModelShading.setRainbowLineSegmentColors(elbow1_p[currentArm].model);
         ModelShading.setRainbowLineSegmentColors(elbow2_p[currentArm].model);
         ModelShading.setRainbowLineSegmentColors(wrist1_p[currentArm].model);
         ModelShading.setRainbowLineSegmentColors(wrist2_p[currentArm].model);
         ModelShading.setRainbowLineSegmentColors(finger1_p[currentArm].model);
         ModelShading.setRainbowLineSegmentColors(finger2_p[currentArm].model);
      }
      else if ('R' == c)
      {
         // Change the random color at each vertex of the robot arm.
         Color c1 = ModelShading.randomColor();
         Color c2 = ModelShading.randomColor();
         Color c3 = ModelShading.randomColor();
         Color c4 = ModelShading.randomColor();
         Color c5 = ModelShading.randomColor();
         Color c6 = ModelShading.randomColor();
         Color c7 = ModelShading.randomColor();
         Color c8 = ModelShading.randomColor();
             arm_p[currentArm].model.colorList = java.util.Arrays.asList(c1, c2);
          elbow1_p[currentArm].model.colorList = java.util.Arrays.asList(c2, c3);
          elbow2_p[currentArm].model.colorList = java.util.Arrays.asList(c2, c4);
          wrist1_p[currentArm].model.colorList = java.util.Arrays.asList(c3, c5);
          wrist2_p[currentArm].model.colorList = java.util.Arrays.asList(c4, c6);
         finger1_p[currentArm].model.colorList = java.util.Arrays.asList(c5, c7);
         finger2_p[currentArm].model.colorList = java.util.Arrays.asList(c6, c8);
             arm_p[currentArm].model.getLineSegment(0).setColors(0, 1);
          elbow1_p[currentArm].model.getLineSegment(0).setColors(0, 1);
          elbow2_p[currentArm].model.getLineSegment(0).setColors(0, 1);
          wrist1_p[currentArm].model.getLineSegment(0).setColors(0, 1);
          wrist2_p[currentArm].model.getLineSegment(0).setColors(0, 1);
         finger1_p[currentArm].model.getLineSegment(0).setColors(0, 1);
         finger2_p[currentArm].model.getLineSegment(0).setColors(0, 1);
      }
      else if ('=' == c)
      {
         xTranslation[currentArm] =  0.0;
         if (0 == currentArm)
            yTranslation[0] =  0.5;
         else
            yTranslation[1] = -0.5;

        shoulderRotation[currentArm] = 0.0;
           elbowRotation1[currentArm] =  15.0;
           elbowRotation2[currentArm] = -15.0;
           wristRotation1[currentArm] = 0.0;
           wristRotation2[currentArm] = 0.0;
          fingerRotation1[currentArm] = 0.0;
          fingerRotation2[currentArm] = 0.0;

           shoulderLength[currentArm] = 0.4;
             elbowLength1[currentArm] = 0.3;
             elbowLength2[currentArm] = 0.3;
             wristLength1[currentArm] = 0.2;
             wristLength2[currentArm] = 0.2;
            fingerLength1[currentArm] = 0.1;
            fingerLength2[currentArm] = 0.1;
      }
      else if ('x' == c)
      {
         xTranslation[currentArm] += 0.02;
      }
      else if ('X' == c)
      {
         xTranslation[currentArm] -= 0.02;
      }
      else if ('y' == c)
      {
         yTranslation[currentArm] += 0.02;
      }
      else if ('Y' == c)
      {
         yTranslation[currentArm] -= 0.02;
      }
      else if ('s' == c)
      {
         shoulderRotation[currentArm] += 2.0;
      }
      else if ('S' == c)
      {
         shoulderRotation[currentArm] -= 2.0;
      }
      else if ('e' == c)
      {
         elbowRotation1[currentArm] += 2.0;
      }
      else if ('E' == c)
      {
         elbowRotation1[currentArm] -= 2.0;
      }
      else if ('w' == c)
      {
         wristRotation1[currentArm] += 2.0;
      }
      else if ('W' == c)
      {
         wristRotation1[currentArm] -= 2.0;
      }
      else if ('f' == c)
      {
         fingerRotation1[currentArm] += 2.0;
      }
      else if ('F' == c)
      {
         fingerRotation1[currentArm] -= 2.0;
      }
      else if ('q' == c)
      {
         elbowRotation2[currentArm] += 2.0;
      }
      else if ('Q' == c)
      {
         elbowRotation2[currentArm] -= 2.0;
      }
      else if ('z' == c)
      {
         wristRotation2[currentArm] += 2.0;
      }
      else if ('Z' == c)
      {
         wristRotation2[currentArm] -= 2.0;
      }
      else if ('d' == c)
      {
         fingerRotation2[currentArm] += 2.0;
      }
      else if ('D' == c)
      {
         fingerRotation2[currentArm] -= 2.0;
      }

      // Update the nested matrices for the sub models.
      arm_p[currentArm].matrix2Identity();
      arm_p[currentArm].matrix.mult(Matrix.translate(xTranslation[currentArm],
                                                     yTranslation[currentArm],
                                                     -1));
      arm_p[currentArm].matrix.mult(Matrix.rotateZ(shoulderRotation[currentArm]));
      arm_p[currentArm].matrix.mult(Matrix.scale(shoulderLength[currentArm],
                                                 shoulderLength[currentArm],
                                                 1));

      elbow1_p[currentArm].matrix2Identity();
      elbow1_p[currentArm].matrix.mult(Matrix.translate(1, 0, 0));
      elbow1_p[currentArm].matrix.mult(Matrix.rotateZ(elbowRotation1[currentArm]));
      elbow1_p[currentArm].matrix.mult(Matrix.scale(elbowLength1[currentArm]/shoulderLength[currentArm],
                                                    elbowLength1[currentArm]/shoulderLength[currentArm],
                                                    1));

      elbow2_p[currentArm].matrix2Identity();
      elbow2_p[currentArm].matrix.mult(Matrix.translate(1, 0, 0));
      elbow2_p[currentArm].matrix.mult(Matrix.rotateZ(elbowRotation2[currentArm]));
      elbow2_p[currentArm].matrix.mult(Matrix.scale(elbowLength2[currentArm]/shoulderLength[currentArm],
                                                    elbowLength2[currentArm]/shoulderLength[currentArm],
                                                    1));

      wrist1_p[currentArm].matrix2Identity();
      wrist1_p[currentArm].matrix.mult(Matrix.translate(1, 0, 0));
      wrist1_p[currentArm].matrix.mult(Matrix.rotateZ(wristRotation1[currentArm]));
      wrist1_p[currentArm].matrix.mult(Matrix.scale(wristLength1[currentArm]/elbowLength1[currentArm],
                                                    wristLength1[currentArm]/elbowLength1[currentArm],
                                                    1));

      wrist2_p[currentArm].matrix2Identity();
      wrist2_p[currentArm].matrix.mult(Matrix.translate(1, 0, 0));
      wrist2_p[currentArm].matrix.mult(Matrix.rotateZ(wristRotation2[currentArm]));
      wrist2_p[currentArm].matrix.mult(Matrix.scale(wristLength2[currentArm]/elbowLength2[currentArm],
                                                    wristLength2[currentArm]/elbowLength2[currentArm],
                                                    1));

      finger1_p[currentArm].matrix2Identity();
      finger1_p[currentArm].matrix.mult(Matrix.translate(1, 0, 0));
      finger1_p[currentArm].matrix.mult(Matrix.rotateZ(fingerRotation1[currentArm]));
      finger1_p[currentArm].matrix.mult(Matrix.scale(fingerLength1[currentArm]/wristLength1[currentArm],
                                                     fingerLength1[currentArm]/wristLength1[currentArm],
                                                     1));

      finger2_p[currentArm].matrix2Identity();
      finger2_p[currentArm].matrix.mult(Matrix.translate(1, 0, 0));
      finger2_p[currentArm].matrix.mult(Matrix.rotateZ(fingerRotation2[currentArm]));
      finger2_p[currentArm].matrix.mult(Matrix.scale(fingerLength2[currentArm]/wristLength2[currentArm],
                                                     fingerLength2[currentArm]/wristLength2[currentArm],
                                                     1));

      // Render again.
      final FrameBuffer fb = this.fbp.getFrameBuffer();
      fb.clearFB();
      Pipeline.render(scene, fb.vp);
      fbp.update();
   }


   // Implement the ComponentListener interface.
   @Override public void componentMoved(ComponentEvent e){}
   @Override public void componentHidden(ComponentEvent e){}
   @Override public void componentShown(ComponentEvent e){}
   @Override public void componentResized(ComponentEvent e)
   {
      //System.out.println( e );
      /*
      System.out.printf("JFrame [w = %d, h = %d]: " +
                        "FrameBufferPanel [w = %d, h = %d].\n",
                        jf.getWidth(), jf.getHeight(),
                        fbp.getWidth(), fbp.getHeight());
      */
      // Get the new size of the FrameBufferPanel.
      final int w = fbp.getWidth();
      final int h = fbp.getHeight();

      // Create a new FrameBuffer that fits the FrameBufferPanel.
      final Color bg1 = fbp.getFrameBuffer().getBackgroundColorFB();
      final Color bg2 = fbp.getFrameBuffer().getViewport().getBackgroundColorVP();
      final FrameBuffer fb = new FrameBuffer(w, h, bg1);
      fb.vp.setBackgroundColorVP(bg2);
      fbp.setFrameBuffer(fb);
      fb.clearFB();
      Pipeline.render(scene, fb.vp);
      fbp.update();
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      print_help_message();

      // We need to call the program's constructor in the
      // Java GUI Event Dispatch Thread, otherwise we get a
      // race condition between the constructor (running in
      // the main() thread) and the very first ComponentEvent
      // (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         () -> new RobotArms3()
      );
   }


   private static void print_help_message()
   {
      //System.out.println("Use the 'd' key to toggle debugging information on and off.");
      //System.out.println("Use the 'a' key to toggle antialiasing on and off.");
      //System.out.println("Use the 'g' key to toggle gamma correction on and off.");
      System.out.println("Use the '/' key to toggle between the the two robot arms.");
      System.out.println();
      System.out.println("Use the 'c' key to change the random solid arm color.");
      System.out.println("Use the 'C' key to randomly change arm segment colors.");
      System.out.println("Use the 'r' key to randomly change arm segment end colors.");
      System.out.println("Use the 'R' key to randomly change arm hinge colors.");
      System.out.println();
      System.out.println("Use the s/S keys to rotate the current arm at the shoulder.");
      System.out.println();
      System.out.println("Use the e/E keys to rotate the current arm at elbow 1.");
      System.out.println("Use the w/W keys to rotate the current arm at wrist 1.");
      System.out.println("Use the f/F keys to rotate the current arm at finger 1.");
      System.out.println();
      System.out.println("Use the q/Q keys to rotate the current arm at elbow 2.");
      System.out.println("Use the z/Z keys to rotate the current arm at wrist 2.");
      System.out.println("Use the d/D keys to rotate the current arm at finger 2.");
      System.out.println();
      System.out.println("Use the ^s/^S keys to extend the length of the current arm at the shoulder.");
      System.out.println();
      System.out.println("Use the ^e/^E keys to extend the length of the current arm at elbow 1.");
      System.out.println("Use the ^w/^W keys to extend the length of the current arm at wrist 1.");
      System.out.println("Use the ^f/^F keys to extend the length of the current arm at finger 1.");
      System.out.println();
      System.out.println("Use the ^q/^Q keys to extend the length of the current arm at elbow 2.");
      System.out.println("Use the ^z/^Z keys to extend the length of the current arm at wrist 2.");
      System.out.println("Use the ^d/^D keys to extend the length of the current arm at finger 2.");
      System.out.println();
      System.out.println("Use the x/X keys to translate the current arm along the x-axis.");
      System.out.println("Use the y/Y keys to translate the current arm along the y-axis.");
      System.out.println();
      System.out.println("Use the '=' key to reset the current robot arm.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }
}