/*

*/

import renderer.scene.*;
import renderer.models.Sphere;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;
import javax.swing.JFrame;
import java.awt.BorderLayout;
import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;
import java.awt.event.ComponentListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.Timer;

/**
   Draw an animation of a solar system with a sun, planet, and moon.
<p>
   In this version, the planet orbits around the sun at the same rate as the sun
   rotates on its axis. Similarly, the moon orbits around the planet at the same
   rate as the planet rotates on its axis.
<p>
   But the orbit of the planet should be independent of the rotation of the sun,
   and the orbit of the moon should be independent of the rotation of the planet.
<pre>{@code
           Scene
             |
             |
          Position
         /   |     \
        /    |      \
  Matrix   Model    nested Positions
    R      (sun)          |
                          |
                       Position
                      /   |   \
                     /    |    \
               Matrix   Model   nested Positions
                 TR    (planet)      |
                                     |
                                  Position
                                 /   |    \
                                /    |     \
                            Matrix  Model   nested Positions
                              TR   (moon)       |
                                                |
                                              empty
}</pre>
*/
public class SolarSystem implements ActionListener, KeyListener, ComponentListener
{
   private int fps;
   private Timer timer = null;

   private double planetOrbitRadius = 5.0;
   private double   moonOrbitRadius = 1.0;

   private double planetOrbitRot = 0.0;
   private double   moonOrbitRot = 0.0;

   private double  moonAxisRot = 0.0;

   private double ecliptic = 7.0; // angle of the ecliptic plane

   private Scene scene = null;
   private Position sun_p = null;
   private Position planetMoon_p = null;
   private Position moon_p = null;

   private JFrame jf = null;
   private FrameBufferPanel fbp = null;

   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
      Then this constructor instantiates the GUI.
   */
   public SolarSystem()
   {
      /*
         See the above picture of the tree that this code creates.
      */

      // Create the Scene object that we shall render
      scene = new Scene();  // A Solar System

      // Create the sun.
      scene.addPosition(new Position(new Sphere(1.0, 10, 10)));
      ModelShading.setColor(scene.getPosition(0).model, Color.yellow);
      sun_p = scene.getPosition(0);

      // Create the planet.
      sun_p.addNestedPosition(new Position(new Sphere(0.5, 10, 10)));
      ModelShading.setColor(sun_p.getNestedPosition(0).model, Color.blue);
      planetMoon_p = sun_p.getNestedPosition(0);

      // Create the planet's moon.
      planetMoon_p.addNestedPosition(new Position(new Sphere(0.2, 10, 10)));
      ModelShading.setColor(planetMoon_p.getNestedPosition(0).model, Color.green);
      moon_p = planetMoon_p.getNestedPosition(0);


      // Define initial dimensions for a FrameBuffer.
      final int width  = 1024;
      final int height = 1024;

      // Create a FrameBufferPanel that will hold a FrameBuffer.
      fbp = new FrameBufferPanel(width, height);
      fbp.getFrameBuffer().setBackgroundColorFB(Color.black);
      fbp.getFrameBuffer().getViewport().setBackgroundColorVP(Color.black);

      // Create a JFrame that will hold the FrameBufferPanel.
      jf = new JFrame("Renderer 8");
      jf.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    //jf.setLocationByPlatform(true);
      // Place the FrameBufferPanel in the JFrame.
      jf.getContentPane().add(fbp, BorderLayout.CENTER);
      jf.pack();
      jf.setVisible(true);

      // Create event handler objects for events from the JFrame.
      jf.addKeyListener(this);
      jf.addComponentListener(this);

      fps = 20;
      timer = new Timer(1000/fps, this); // ActionListener
      timer.start();
   }


   // Implement the ActionListener interface.
   @Override public void actionPerformed(ActionEvent e)
   {
      //System.out.println( e );

      // Push the solar system away from where the camera is.
      sun_p.matrix2Identity();
      sun_p.matrix.mult(Matrix.translate(0, 0, -8));
      // Rotate the plane of the ecliptic
      // (rotate the sun's xz-plane about the x-axis).
      sun_p.matrix.mult(Matrix.rotateX(ecliptic));

      // Rotate the sun on it axis.
      sun_p.matrix.mult(Matrix.rotateY(planetOrbitRot));

      // Place the planet-moon away from the sun and rotate the planet-moon on its axis.
      planetMoon_p.matrix2Identity();
      planetMoon_p.matrix.mult(Matrix.translate(planetOrbitRadius, 0, 0));
      planetMoon_p.matrix.mult(Matrix.rotateY(moonOrbitRot));

      // Place the moon away from the planet and rotate the moon on its axis.
      moon_p.matrix2Identity();
      moon_p.matrix.mult(Matrix.translate(moonOrbitRadius, 0, 0));
      moon_p.matrix.mult(Matrix.rotateY(moonAxisRot));

      // Update the parameters for the next frame.
      planetOrbitRot += 1.0;
      moonOrbitRot += 5.0;
      moonAxisRot -= 10.0;

      // Render again.
      final FrameBuffer fb = this.fbp.getFrameBuffer();
      fb.clearFB();
      Pipeline.render(scene, fb.vp);
      fbp.update();
   }


   // Implement the KeyListener interface.
   @Override public void keyPressed(KeyEvent e){}
   @Override public void keyReleased(KeyEvent e){}
   @Override public void keyTyped(KeyEvent e)
   {
      //System.out.println( e );

      final char c = e.getKeyChar();
      if ('h' == c)
      {
         print_help_message();
         return;
      }
      else if ('d' == c)
      {
         Pipeline.debug = ! Pipeline.debug;
         Clip.debug = ! Clip.debug;
      }
      else if ('D' == c)
      {
         RasterizeAntialias.debug = ! RasterizeAntialias.debug;
      }
      else if ('a' == c)
      {
         RasterizeAntialias.doAntialiasing = ! RasterizeAntialias.doAntialiasing;
         System.out.print("Anti-aliasing is turned ");
         System.out.println(RasterizeAntialias.doAntialiasing ? "On" : "Off");
      }
      else if ('g' == c)
      {
         RasterizeAntialias.doGamma = ! RasterizeAntialias.doGamma;
         System.out.print("Gamma correction is turned ");
         System.out.println(RasterizeAntialias.doGamma ? "On" : "Off");
      }
      else if ('f' == c)
      {
         fps -= 1;
         if (0 > fps) fps = 0;
         System.out.println("fps = " + fps);
         setFPS();
      }
      else if ('F' == c)
      {
         fps += 1;
         System.out.println("fps = " + fps);
         setFPS();
      }
      else if ('p' == c)
      {
         scene.camera.perspective = ! scene.camera.perspective;
         final String p = scene.camera.perspective ? "perspective" : "orthographic";
         System.out.println("Using " + p + " projection");
      }
      else if ('e' == c)
      {
         ecliptic -= 1;
         System.out.println("ecliptic = " + ecliptic);
      }
      else if ('E' == c)
      {
         ecliptic += 1;
         System.out.println("ecliptic = " + ecliptic);
      }

      // Set up the camera's view volume.
      if (scene.camera.perspective)
      {
         scene.camera.projPerspective(-1, 1, -1, 1, 1);
      }
      else
      {
         scene.camera.projOrtho(-6, 6, -6, 6);
      }
   }


   public void setFPS()
   {
      timer.stop();
      if (fps > 0)
      {
         timer = new Timer(1000/fps, this);
         timer.start();
      }
   }


   // Implement the ComponentListener interface.
   @Override public void componentMoved(ComponentEvent e){}
   @Override public void componentHidden(ComponentEvent e){}
   @Override public void componentShown(ComponentEvent e){}
   @Override public void componentResized(ComponentEvent e)
   {
      //System.out.println( e );
      /*
      System.out.printf("JFrame [w = %d, h = %d]: " +
                        "FrameBufferPanel [w = %d, h = %d].\n",
                        jf.getWidth(), jf.getHeight(),
                        fbp.getWidth(), fbp.getHeight());
      */
      // Get the new size of the FrameBufferPanel.
      final int w = fbp.getWidth();
      final int h = fbp.getHeight();

      // Create a new FrameBuffer that fits the FrameBufferPanel.
      final Color bg1 = fbp.getFrameBuffer().getBackgroundColorFB();
      final Color bg2 = fbp.getFrameBuffer().getViewport().getBackgroundColorVP();
      final FrameBuffer fb = new FrameBuffer(w, h, bg1);
      fb.vp.setBackgroundColorVP(bg2);
      fbp.setFrameBuffer(fb);
      fb.clearFB();
      Pipeline.render(scene, fb.vp);
      fbp.update();
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      print_help_message();

      // We need to call the program's constructor in the
      // Java GUI Event Dispatch Thread, otherwise we get a
      // race condition between the constructor (running in
      // the main() thread) and the very first ComponentEvent
      // (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         () -> new SolarSystem()
      );
   }


   private static void print_help_message()
   {
      System.out.println("Use the 'd' key to toggle debugging information on and off.");
      System.out.println("Use the 'a' key to toggle antialiasing on and off.");
      System.out.println("Use the 'g' key to toggle gamma correction on and off.");
      System.out.println("Use the f/F keys to slow down or speed up the frame rate.");
      System.out.println("Use the 'p' key to toggle between parallel and orthographic projection.");
      System.out.println("Use the e/E keys to change the angle of the ecliptic plane.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }
}
